import inspect
import re
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, Optional, Type, get_origin

from .objectid_proxy import ObjectId

Recur = Callable[[Any, Type], Any]
Serializer = Callable[[Any, Type, Recur], Any]
Deserializer = Callable[[Any, Type, Recur], Any]


_SPLIT_CAMEL_CASE_PATTERN = re.compile(
    ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


class SerdeError(Exception):
    """
    Signals an error during serialization or deserialization.
    """

    def __init__(self, user_message: str):
        self.user_message = user_message


def as_child(cls: Type) -> Type:
    """
    Marks the class to be serialized as one of its children. This will add an
    additional "type" field in the result, so the child can be deserialized
    properly.

    This decorator applies to children of the class as well, i.e. they will also
    be serialized with the "type" field.
    """
    assert inspect.isclass(cls), cls
    cls.__serde_serialize_as_child__ = cls  # type: ignore
    return cls


def should_serialize_as_child(cls: Type) -> bool:
    """
    Checks whether the given class should be serialized as a child, i.e. it, or
    any parent has been marked with the `as_child` decorator.
    """
    assert inspect.isclass(cls), cls
    return hasattr(cls, "__serde_serialize_as_child__")


def common_serialize(
    value: Any,
    value_type: Optional[Type],
    class_serializer: Serializer,
    serializers: Dict[Type, Serializer],
    user_serializers: Dict[Type, Callable[[Any], Any]],
) -> Any:
    # Find the type
    if value_type is None:
        value_type = type(value)

    # Is there a custom serializer for this class?
    try:
        serializer = user_serializers[value_type]
    except KeyError:
        pass
    else:
        return serializer(value)

    # Find a matching serializer
    key = get_origin(value_type)
    if key is None:
        key = value_type

    try:
        serializer = serializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            serializer = class_serializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_serialize(
            value,
            value_type,
            class_serializer,
            serializers,
            user_serializers,
        )

    # Apply it
    return serializer(value, value_type, recur)


def common_deserialize(
    value: Any,
    value_type: Type,
    class_deserializer: Serializer,
    deserializers: Dict[Type, Deserializer],
    user_deserializers: Dict[Type, Callable[[Any], Any]],
) -> Any:
    # Is there a custom deserializer for this class?
    try:
        deserializer = user_deserializers[value_type]
    except KeyError:
        pass
    else:
        return deserializer(value)

    # Find a matching deserializer
    key = get_origin(value_type)
    if key is None:
        key = value_type

    try:
        deserializer = deserializers[key]
    except KeyError:
        if inspect.isclass(value_type):
            deserializer = class_deserializer
        else:
            raise ValueError(f"Unsupported field of type {value_type}") from None

    # Define the recursion function
    def recur(value: Any, value_type: Type) -> Any:
        return common_deserialize(
            value,
            value_type,
            class_deserializer,
            deserializers,
            user_deserializers,
        )

    # Apply it
    return deserializer(value, value_type, recur)


def all_subclasses(cls: Type, include_cls: bool) -> Iterable[Type]:
    """
    Yields all classes directly on indirectly inheriting from `cls`. Does not
    perform any sort of cycle checks.

    :param cls: The class to start from.
    :param include_cls: Whether to include `cls` itself in the results.
    """

    if include_cls:
        yield cls

    for subclass in cls.__subclasses__():
        yield from all_subclasses(subclass, include_cls=True)



import enum
import inspect
from abc import ABC, abstractmethod
from typing import *  # type: ignore

from typing_extensions import Self

import uniserde

from . import case_convert, common, serde_class
from .common import SerdeError

T = TypeVar("T")
V = TypeVar("V")


class CachingDeserializer(ABC, Generic[T]):
    _passthrough_types: Set[Type]
    _deserializer_cache: Dict[Type, Callable[[Self, T, Type], Any]]
    _override_method_name: str

    def __init_subclass__(cls) -> None:
        def make_deserializer_from_passthrough_type(
            passthrough_type: Type,
        ) -> Callable[[Self, T, Type], Any]:
            def result(self, value: T, as_type: Type[V]) -> V:
                if not isinstance(value, as_type):
                    raise SerdeError(f"Expected `{as_type}`, got `{value}`")

                return value

            return result

        for typ in cls._passthrough_types:
            cls._deserializer_cache[typ] = make_deserializer_from_passthrough_type(typ)

    def __init__(
        self,
        *,
        custom_deserializers: Dict[Type, Callable[[T, Type], Any]] = {},
    ):
        self._custom_deserializers = custom_deserializers

    @abstractmethod
    def _get_class_fields_and_deserializers(
        self, value_type: Type
    ) -> Iterable[Tuple[str, str, Callable[[T, Type], Any]]]:
        """
        Return a list of (python_name, json_name, deserializer) tuples for each
        field in the class.
        """
        raise NotImplementedError()

    def deserialize(self, value: T, as_type: Type[V]) -> V:
        # Special case: passthrough types for performance
        if as_type in self._passthrough_types:
            if not isinstance(value, as_type) and not (
                isinstance(value, int) and as_type is float
            ):
                raise SerdeError(f"Expected `{as_type}`, got `{value}`")

            return value  # type: ignore

        # Otherwise get a deserializer and use it
        deserializer = self._get_deserializer(as_type)
        return deserializer(self, value, as_type)

    def _get_deserializer(
        self,
        value_type: Type,
    ) -> Callable[[Self, T, Type], Any]:
        # Prepare the key for the deserializer lookup
        key = get_origin(value_type)
        if key is None:
            key = value_type

        # Custom deserializers take precedence
        try:
            custom_callback = self._custom_deserializers[key]
        except KeyError:
            pass
        else:
            return lambda self, value, as_type: custom_callback(value, as_type)

        # Use a cached deserializer if possible
        try:
            return self._deserializer_cache[key]
        except KeyError:
            pass

        # Otherwise create the appropriate deserializer and cache it for next
        # time
        assert inspect.isclass(value_type), value_type
        deserializer = self._create_class_deserializer(value_type)
        self._deserializer_cache[key] = deserializer

        return deserializer

    def _create_class_deserializer(
        self,
        value_type: Type,
    ) -> Callable[[Self, T, Type], Any]:
        # Case: The class has a custom deserialization method
        try:
            override_method = getattr(value_type, self._override_method_name)
        except AttributeError:
            pass
        else:
            serde_class_method = getattr(serde_class.Serde, self._override_method_name)

            if override_method.__func__ is not serde_class_method.__func__:
                return lambda self, value, _type: override_method(value, {})

        # Case: Enum
        if issubclass(value_type, enum.Enum):

            def deserialize_enum(self, value, _type):
                if not isinstance(value, str):
                    raise SerdeError(f"Expected enumeration string, got `{value}`")

                try:
                    py_name = case_convert.camel_case_to_all_upper(
                        value
                    )  # ValueError if not camel case
                    return value_type[py_name]  # ValueError if not in enum
                except KeyError:
                    raise SerdeError(f"Invalid enumeration value `{value}`") from None

            return deserialize_enum

        # Case: Base which is serialized `@as_child`
        if common.should_serialize_as_child(value_type):
            # Prepare a set of all possible classes
            child_classes_and_deserializers_by_doc_name = {
                case_convert.upper_camel_case_to_camel_case(cls.__name__): (
                    cls,
                    self._create_fieldwise_class_deserializer(cls),
                )
                for cls in common.all_subclasses(value_type, True)
            }

            def deserialize_as_child(self, value, _type):
                # Look up the real type
                try:
                    type_tag = value.pop("type")
                except KeyError:
                    raise SerdeError(f"Object is missing the `type` field") from None

                # Get the class
                try:
                    (
                        child_class,
                        child_class_deserializer,
                    ) = child_classes_and_deserializers_by_doc_name[type_tag]
                except KeyError:
                    raise SerdeError(
                        f"Encountered invalid type tag `{type_tag}`"
                    ) from None

                # Delegate to that class's deserializer
                return child_class_deserializer(self, value, child_class)

            return deserialize_as_child

        # Case: Regular old class
        return self._create_fieldwise_class_deserializer(value_type)

    def _create_fieldwise_class_deserializer(
        self, value_type: Type
    ) -> Callable[[Self, T, Type], Any]:
        deserializer = FieldwiseClassDeserializer()

        for py_name, doc_name, field_type in self._get_class_fields_and_deserializers(
            value_type
        ):
            deserializer.add_field(
                py_name,
                doc_name,
                field_type,
                self._get_deserializer(field_type),
            )

        return deserializer


class FieldwiseClassDeserializer:
    fields: List[Tuple[str, str, Type, Callable[[CachingDeserializer, Any, Type], Any]]]

    def __init__(self):
        self.fields = []

    def add_field(
        self,
        python_name: str,
        doc_name: str,
        field_type: Type,
        deserializer: Callable[[CachingDeserializer, Any, Type], Any],
    ):
        self.fields.append((python_name, doc_name, field_type, deserializer))

    def __call__(
        self,
        calling_deserializer: CachingDeserializer,
        raw: Any,
        value_type: Type,
    ) -> Any:
        # Make sure the raw value is a dict
        if not isinstance(raw, dict):
            raise uniserde.SerdeError(f"Expected object, got `{raw!r}`")

        # Create an instance of the class
        result = object.__new__(value_type)
        result_dict = vars(result)

        # Deserialize all fields
        for py_name, doc_name, field_type, deserializer in self.fields:
            # Get the raw value
            try:
                raw_value = raw.pop(doc_name)
            except KeyError:
                raise uniserde.SerdeError(f"Missing field `{doc_name!r}`") from None

            # Deserialize it
            parsed_value = deserializer(calling_deserializer, raw_value, field_type)

            # Store it
            result_dict[py_name] = parsed_value

        # Make sure there are no stray fields
        if len(raw) > 0:
            raise SerdeError(
                f"Superfluous object fields `{'`, `'.join(map(str, raw.keys()))}`"
            )

        return result



from datetime import datetime
from typing import *  # type: ignore

from .objectid_proxy import ObjectId

__all__ = [
    "Jsonable",
    "Bsonable",
]


Jsonable = Union[
    None,
    bool,
    int,
    float,
    str,
    Dict[str, "Jsonable"],
    List["Jsonable"],
    Tuple["Jsonable", ...],
]


Bsonable = Union[
    None,
    bool,
    int,
    float,
    str,
    Dict[str, "Bsonable"],
    List["Bsonable"],
    Tuple["Bsonable", ...],
    bytes,
    ObjectId,
    datetime,
]



from datetime import datetime
from typing import *  # type: ignore

from uniserde.objectid_proxy import ObjectId

from . import caching_deserializer, case_convert, json_deserialize
from .typedefs import Bsonable

__all__ = [
    "from_bson",
]


T = TypeVar("T")


class BsonDeserializer(caching_deserializer.CachingDeserializer[Bsonable]):
    def _get_class_fields_and_deserializers(
        self, value_type: Type
    ) -> Iterable[Tuple[str, str, Callable[[Bsonable, Type], Any]]]:
        for field_name_py, field_type in get_type_hints(value_type).items():
            if field_name_py == "id":
                field_name_doc = "_id"
            else:
                field_name_doc = case_convert.all_lower_to_camel_case(field_name_py)

            yield (
                field_name_py,
                field_name_doc,
                field_type,
            )

    _passthrough_types = {
        bool,
        int,
        float,
        bytes,
        str,
        datetime,
        ObjectId,
    }
    _deserializer_cache = json_deserialize.JsonDeserializer._deserializer_cache.copy()  # type: ignore
    _override_method_name = "from_bson"


def from_bson(
    value: Any,
    as_type: Type[T],
    *,
    custom_deserializers: Dict[Type, Callable[[Bsonable], Any]] = {},
) -> T:
    deserializer = BsonDeserializer(
        custom_deserializers={
            t: lambda v, _: cb(v) for t, cb in custom_deserializers.items()
        }
    )

    return deserializer.deserialize(value, as_type)



from .bson_deserialize import from_bson
from .common import SerdeError, as_child
from .json_deserialize import from_json
from .objectid_proxy import ObjectId
from .schema_mongodb import as_mongodb_schema
from .serde_bson import *
from .serde_class import Serde
from .serde_json import *
from .typedefs import Bsonable, Jsonable
